#include <iostream>
#include <fstream>
#include <string>
#include <math.h>

using namespace std;

// globale Variablen
const int N = 300;
const double h = 1./double(N-1);
const int MAX_ITER = N*N;
const double TOL = 1e-10;

// schreibe als ppm
void writePPM(const char* filename, const double data[][N]) {
	cout<<"Saving to file "<<filename<<endl;
	double min = data[0][0];
	double max = data[0][0];
  for( int i = 0; i < N; i++ )
    for( int j = 0; j < N; j++ ) {
		  if (data[i][j]<min) min=data[i][j];
		  if (data[i][j]>max) max=data[i][j];
    }

  ofstream out ( filename );
  out << "P3" << endl;
  out << N << " " << N << endl;
  out << 255 << endl;
  for( int i = 0; i < N; i++ )
    for( int j = 0; j < N; j++ ) {
    	int col = 255*(data[j][N-1-i]-min)/(max-min);
    	out << col << " " << col << " " << col << endl;
    }
}

// Skalar Produkt
double dot(const double chi[][N], const double x[][N], const double y[][N]) {
  double norm = 0.;
  for( int i = 0; i < N; i++ ) 
    for( int j = 0; j < N; j++ )
      if (chi[i][j]>0.5)
        norm += x[i][j] * y[i][j];
      return norm;
}


// Anwendung der Matrix L auf den Vecktor U,
// wobei L die finite Differenzen Diskretisierung des (negativen) Laplace-Operators ist.
// Keine Assemblierung der Matrix, sondern iterative Anwendung des Differenzensterns
//         |    -1     |
//  h^{-2} | -1  4  -1 |
//         |    -1     |
void applyMatrix(const double U[][N], const double chi[][N], double LU[][N] ) {
 
  const double h2 = h*h;

  for( int i = 0; i < N; i++ )
    for( int j = 0; j < N; j++ )
      if (chi[i][j]>0.5) {
       LU[i][j] = (4*U[i][j] - U[i-1][j] - U[i+1][j] - U[i][j-1] - U[i][j+1])/h2;
      } 
      else LU[i][j] = 0.;
}


// Konjugierte Gradienten Methode zur Loesung des Systems Lx=b,
// wobei L die finite Differenzen Diskretisierung des (negativen) Laplace-Operators ist.
void applyCG(const double b[][N], double x[][N], const double chi[][N]) {
  static double r[N][N];
  static double p[N][N];
  static double Lp[N][N];

  // Erste Suchrichtung p = r_0
  // mit Residuum r_0 = b - L * x
  applyMatrix( x, chi, Lp );
  for( int i = 0; i < N; i++ ) 
    for( int j = 0; j < N; j++ )
      if (chi[i][j]>0.5) {
        r[i][j] = b[i][j] - Lp[i][j];
        p[i][j] = r[i][j];                   
      } 

  // berechne Norm des Residuums und Fehler
  double norm_rq = dot(chi,r,r);
  double error = sqrt( norm_rq ); 
  
  int iter = 1;

  // Iteriere bis die Fehlerschranke unterschritten bzw die 
  // maximale Iterationszahl erreicht ist.
  while ( ( iter < MAX_ITER) && ( error > TOL ) ) {
    // wende Matrix an
    applyMatrix( p, chi, Lp );
	
    // alpha = Norm(r)^2 / (p^T * L * p)
    double pLp = dot(chi,p,Lp);
    double alpha = norm_rq / pLp;       

    // Berechne Residuum r und Position x neu
    for( int i = 0; i < N; i++ ) 
      for( int j = 0; j < N; j++ )
        if (chi[i][j]>0.5) {
          x[i][j] += alpha * p[i][j];   // x = x + alpha * p
          r[i][j] -= alpha * Lp[i][j];  // r = r - alpha * L * p
        }      

    // berechne Norm des Residuums neu
    double norm_rq_alt = norm_rq;
    norm_rq = dot(chi,r,r);

    // Berechnung der Suchrichtung für den nächsten Schritt
    double beta = norm_rq / norm_rq_alt;
    // p_neu = beta * p_alt + r
    for( int i = 0; i < N; i++ ) 
      for( int j = 0; j < N; j++ )
        if (chi[i][j]>0.5)
          p[i][j] = beta * p[i][j] + r[i][j];

    // Aktualisierung des Fehlers
    error = sqrt( norm_rq );
    // Iterationszahl inkrementieren
    iter++;
  }    
  
  // werte das Abbruchverhalten aus
  if ( error > TOL )
    cerr << "CG: Could not solve to required tolerance ( "<< TOL << " ) within " << MAX_ITER << " Iterations !!!" << endl;
  else
    cout << "CG: Solved to required tolerance ( "<< TOL << " ) within " << iter << " Iterations." << endl;
}

// Hauptprogramm
int main () 
{
  // Definiere die Gebietsmaske chi sowie die rechte Seite f = 1
  static double chi[N][N];
  static double f[N][N];
  for( int i = 0; i < N; i++ ) {
    for( int j = 0; j < N; j++ ) {
      double x = i*h; //x,y in [0,1]
      double y = j*h;
      // das L-Gebiet ist gegeben durch \Omega = \{ (x,y) \in [0.1, 0.9]^2 : y > .5 falls x > .5 \}
      chi[i][j] = ((.1<x && x<.9 && .1<y && y<.9) && !(x>.5 && y<.5)) ? 1.0 : 0.0;
      f[i][j] = 1.0;
    }
  }
  
  // schreibe chi und f 
  writePPM("chi.ppm",chi);
  writePPM("f.ppm",f);

  // setze Randbedingungen fuer U
  static double U[N][N];
  for( int i = 0; i < N; i++ ) 
    for( int j = 0; j < N; j++ )
      U[i][j]= 0.;
  writePPM("u0.ppm",U);
  
  // loese Lu = f mit dem CG-Verfahren
  applyCG(f, U, chi);
  // speichere die Loesung
  writePPM("u.ppm",U);

  return 0;
}